# -*- coding: utf-8 -*-

import numpy as np
import scipy.constants
from scipy.optimize import curve_fit, fsolve

KB = scipy.constants.Boltzmann
hbar = scipy.constants.hbar

from lmfit import Minimizer, Parameters


def fit_S11(params, w, data=None):
    w0 = params["w0"]
    k = params["k"]
    kex = params["kex"]

    try:
        directivity = params["directivity"]
    except:
        directivity = 0

    try:
        tau = params["tau"]
    except:
        tau = 0
    try:
        bkg = params["background"]
    except:
        bkg = 0
    try:
        phi0 = params["phi0"]
    except:
        phi0 = 0
    try:
        q = params["q"]
    except:
        q = 0
    try:
        A = params["A"]
    except:
        A = 1
    try:
        B = params["B"]
    except:
        B = 0

    if bkg == 1:
        env = (A + B * w) * np.exp(1j * (phi0 + w * tau))
    else:
        env = 1
        q = 0

    num = -2 * kex + k + 2j * (w - w0) + 2j * q
    denom = k + 2j * (w - w0)
    out = num / denom + directivity

    if data is None:
        return env * out
    else:
        fit = np.concatenate((np.real(env * out), np.imag(env * out)))
        return fit - data


def S11_series(w, C, L, Ls, Rp, RLs):
    Rex = 50
    # C=10e-15
    Zcav = 1 / (1j * w * C) + 1 / (1 / (1j * w * L) + 1 / Rp)

    if Ls is None:
        Z = Zcav
    else:
        # Z=1/(1/Zcav+1/(1j*w*Ls)+1/RLs)
        Z = 1 / (1 / Zcav + 1 / (1j * w * Ls) + 1 / RLs)

    num = Rex - Z
    denom = Rex + Z
    out = num / denom
    # return out*np.exp(-1j*np.angle(out[0]))
    return out * np.exp(-1j * -0.25)


def S11_parallel(w, C, L, Cc, Rp):  # Rp is the parallel of Rd and RL
    Rex = 50
    # C=10e-15

    Zcav = 1 / (1j * w * C + 1 / (1j * w * L) + 1 / Rp)

    Z = 1 / (1j * w * Cc) + Zcav

    num = Rex - Z
    denom = Rex + Z

    S11 = num / denom

    return S11 * (np.exp(-1j * np.angle(S11)[0]))
    # return Z


def S11_parallel_lm(params, w, data=None):  # Rp is the parallel of Rd and RL
    try:
        Rex = params["Rex"]
    except:
        Rex = 50
    C = params["C"]
    L = params["L"]
    Cc = params["Cc"]
    Rp = params["Rp"]

    Zcav = 1 / (1j * w * C + 1 / (1j * w * L) + 1 / Rp)

    Z = 1 / (1j * w * Cc) + Zcav

    num = Rex - Z
    denom = Rex + Z
    out = num / denom
    # out=out*np.exp(1j*np.angle(out[-1]))
    out = out * np.exp(1j * np.pi)

    if data is None:
        return out
    else:
        fit = np.concatenate((np.real(out), np.imag(out)))
        return fit - data


def fL(T, Tc, D0):
    # D0=hbar/np.pi*1e12*Tc
    y = (
        1
        / np.sqrt(np.cos(np.pi / 2 * (T / Tc) ** 2))
        / np.tanh(D0 * np.sqrt(np.cos(np.pi / 2 * (T / Tc) ** 2)) / (2 * KB * T))
    )
    return y


def fR(T, Tc, D0):
    # D0=hbar/np.pi*1e12*Tc
    y = (
        1
        / np.sqrt(np.cos(np.pi / 2 * (T / Tc) ** 2))
        / np.tanh(D0 * np.sqrt(np.cos(np.pi / 2 * (T / Tc) ** 2)) / (2 * KB * T))
    )
    z = 1 / (1 - 1 / y)
    return z


def fL1(T, Tc):
    y = 1 / (1 - np.power(T / Tc, 4))
    return y


def fR1(T, Tc):
    y = np.power(Tc / T, 4)
    return y


def omega_T(T, Tc, k, w0):
    D0 = hbar / np.pi * k * Tc
    fwt = w0 * np.sqrt(1 / fL(T, Tc, D0))
    return fwt


def omega2_T(params, T, data=None):
    C = params["C"]
    Rd = params["Rd"]
    nu = params["nu"]
    Tc = params["Tc"]
    L0 = params["L0"]
    # print(C)
    Ls0 = params["Ls0"]
    RL0 = params["RL0"]
    D0 = hbar / np.pi * nu * Tc
    w00 = 1 / np.sqrt(C * (L0 + Ls0))
    RL = RL0 * fR(T, Tc, D0)
    R = Rd * RL / (Rd + RL)
    L = L0 * fL(T, Tc, D0)
    Ls = Ls0 * fL(T, Tc, D0)

    w0 = 1 / np.sqrt(C * (L * R**2 / (w00**2 * L**2 + R**2) + Ls))

    if data is None:
        return w0
    else:
        return w0 - data


def MB_Data_SeriesLC_Ls(params, T_arr, data=None):  #### MB, BCS
    f0 = params["f0"]
    span = params["f_span"]

    C = params["C"]
    Rd = params["Rd"]
    nu = params["nu"]
    Tc = params["Tc"]
    L0 = params["L0"]
    Ls0 = params["Ls0"]
    RL0 = params["RL0"]

    D0 = hbar / np.pi * nu * Tc

    w0_list = []
    k0_list = []
    kex_list = []
    for j, T in enumerate(T_arr):
        if data is None:
            w = (
                2
                * np.pi
                * (
                    np.arange(-span / 2, span / 2, 1e5)
                    + f0 * np.sqrt(1 - (T / Tc) ** 4)
                )
            )

        else:
            w0_data, k0_data, kext_data = np.split(data, 3)

            w = (
                np.arange(
                    -(kext_data[j] + k0_data[j]) * 5,
                    (kext_data[j] + k0_data[j]) * 5,
                    2 * np.pi * 1e5,
                )
                + w0_data[j]
            )

        RL = RL0 * fR(T, Tc, D0)
        L = L0 * fL(T, Tc, D0)
        Ls = Ls0 * fL(T, Tc, D0)
        RLs0 = Ls0 * RL0 / L0
        RLs = RLs0 * fR(T, Tc, D0)

        Rp = RL * Rd / (RL + Rd)

        Y = S11_series(w, C, L, Ls, Rp, RLs)

        S11_pars2 = Parameters()
        S11_pars2.add("w0", value=w[np.argmin(np.abs(Y))], min=0, vary=True)
        S11_pars2.add("k", value=2 * np.pi * 1e6, min=0, vary=True)
        S11_pars2.add("kex", value=2 * np.pi * 1e6, min=0, vary=True)
        S11_pars2.add("A", value=1, vary=False)
        S11_pars2.add("B", value=0, vary=False)
        S11_pars2.add("phi0", value=0, vary=False)
        S11_pars2.add("q", value=0, vary=False)
        S11_pars2.add("tau", value=0, vary=False)
        S11_pars2.add("background", value=0, vary=False)

        S11_fitter2 = Minimizer(
            fit_S11, S11_pars2, fcn_args=(w, np.concatenate((np.real(Y), np.imag(Y))))
        )
        S11_result2 = S11_fitter2.minimize()

        # fit Y to standard S11 ###
        ### get w0 and kint ###
        w0_list.append(S11_result2.params["w0"].value)
        k0_list.append(S11_result2.params["k"].value - S11_result2.params["kex"].value)
        kex_list.append(S11_result2.params["kex"].value)

    # w0=1/np.sqrt(C*(L*R**2/(w00**2*L**2+R**2)+Ls))
    w0 = np.array(w0_list)
    k0 = np.array(k0_list)
    kex = np.array(kex_list)

    if data is None:
        return np.concatenate((w0, k0, kex))
    else:
        return np.concatenate((w0, k0, kex)) - data


def MB_Data_SeriesLC_Ls_v2(params, T_arr, data=None):  #### MB, BCS
    f0 = params["f0"]
    span = params["f_span"]

    C = params["C"]
    Rd = params["Rd"]
    Rd0 = params["Rd0"]
    beta = params["beta"]

    a = params["a"]
    Tc = params["Tc"]
    Tc2 = params["Tc2"]
    L0 = params["L0"]
    Ls0 = params["Ls0"]

    D0 = a * KB * Tc
    D02 = a * KB * Tc2

    # RL0=D0*L0*np.pi/hbar
    RL0 = D0 * L0 * np.pi / hbar

    w0_list = []
    k0_list = []
    kex_list = []
    for j, T in enumerate(T_arr):
        if data is None:
            w = (
                2
                * np.pi
                * (
                    np.arange(-span / 2, span / 2, 1e5)
                    + f0 * np.sqrt(1 - (T / Tc) ** 4)
                )
            )

        else:
            w0_data, k0_data, kext_data = np.split(data, 3)

            w = (
                np.arange(
                    -(kext_data[j] + k0_data[j]) * 5,
                    (kext_data[j] + k0_data[j]) * 5,
                    2 * np.pi * 1e5,
                )
                + w0_data[j]
            )

        RL = RL0 * fR(T, Tc, D0)
        L = L0 * fL(T, Tc, D0)
        Ls = Ls0 * fL(T, Tc2, D02)
        RLs0 = Ls0 * RL0 / L0
        RLs = RLs0 * fR(T, Tc2, D02)

        # Rp=RL*Rd/(RL+Rd)
        Rp = 1 / (1 / RL + 1 / Rd0 + 1 / (Rd * np.exp(beta / T)))

        Y = S11_series(w, C, L, Ls, Rp, RLs)

        S11_pars2 = Parameters()
        S11_pars2.add("w0", value=w[np.argmin(np.abs(Y))], min=0, vary=True)
        S11_pars2.add("k", value=2 * np.pi * 1e6, min=0, vary=True)
        S11_pars2.add("kex", value=2 * np.pi * 1e6, min=0, vary=True)
        S11_pars2.add("A", value=1, vary=False)
        S11_pars2.add("B", value=0, vary=False)
        S11_pars2.add("phi0", value=0, vary=False)
        S11_pars2.add("q", value=0, vary=False)
        S11_pars2.add("tau", value=0, vary=False)
        S11_pars2.add("background", value=0, vary=False)

        S11_fitter2 = Minimizer(
            fit_S11, S11_pars2, fcn_args=(w, np.concatenate((np.real(Y), np.imag(Y))))
        )
        S11_result2 = S11_fitter2.minimize()

        # fit Y to standard S11 ###
        ### get w0 and kint ###
        w0_list.append(S11_result2.params["w0"].value)
        k0_list.append(S11_result2.params["k"].value - S11_result2.params["kex"].value)
        kex_list.append(S11_result2.params["kex"].value)

    # w0=1/np.sqrt(C*(L*R**2/(w00**2*L**2+R**2)+Ls))
    w0 = np.array(w0_list)
    k0 = np.array(k0_list)
    kex = np.array(kex_list)

    if data is None:
        return np.concatenate((w0, k0, kex))
    else:
        return np.concatenate((w0, k0, kex)) - data


def MB_Data_ParallelLC_Cc(params, T_arr, data=None):  #### MB, BCS
    f0 = params["f0"]
    span = params["f_span"]

    C = params["C"]
    Cc = params["Cc"]
    Rd = params["Rd"]
    nu = params["nu"]
    Tc = params["Tc"]
    L0 = params["L0"]

    RL0 = params["RL0"]

    D0 = hbar / np.pi * nu * Tc

    w0_list = []
    k0_list = []
    kex_list = []
    for j, T in enumerate(T_arr):
        if data is None:
            w = (
                2
                * np.pi
                * (
                    np.arange(-span / 2, span / 2, 1e5)
                    + f0 * np.sqrt(1 - (T / Tc) ** 4)
                )
            )

        else:
            w0_data, k0_data, kext_data = np.split(data, 3)

            w = (
                np.arange(
                    -(kext_data[j] + k0_data[j]) * 5,
                    (kext_data[j] + k0_data[j]) * 5,
                    2 * np.pi * 1e5,
                )
                + w0_data[j]
            )

        RL = RL0 * fR(T, Tc, D0)
        L = L0 * fL(T, Tc, D0)
        Rp = RL * Rd / (RL + Rd)
        Y = S11_parallel(w, C, L, Cc, Rp)

        S11_pars2 = Parameters()
        S11_pars2.add("w0", value=w[np.argmin(np.abs(Y))], min=0, vary=True)
        S11_pars2.add("k", value=2 * np.pi * 1e6, min=0, vary=True)
        S11_pars2.add("kex", value=2 * np.pi * 0.7e6, min=0, vary=True)
        S11_pars2.add("A", value=1, vary=False)
        S11_pars2.add("B", value=0, vary=False)
        S11_pars2.add("phi0", value=0, vary=False)
        S11_pars2.add("q", value=0, vary=False)
        S11_pars2.add("tau", value=0, vary=False)
        S11_pars2.add("background", value=0, vary=False)

        S11_fitter2 = Minimizer(
            fit_S11, S11_pars2, fcn_args=(w, np.concatenate((np.real(Y), np.imag(Y))))
        )
        S11_result2 = S11_fitter2.minimize()

        # fit Y to standard S11 ###
        ### get w0 and kint ###
        w0_list.append(S11_result2.params["w0"].value)
        k0_list.append(S11_result2.params["k"].value - S11_result2.params["kex"].value)
        kex_list.append(S11_result2.params["kex"].value)

    # w0=1/np.sqrt(C*(L*R**2/(w00**2*L**2+R**2)+Ls))
    w0 = np.array(w0_list)
    k0 = np.array(k0_list)
    kex = np.array(kex_list)

    if data is None:
        return np.concatenate((w0, k0, kex))
    else:
        return np.concatenate((w0, k0, kex)) - data


def omega4_T(params, T_arr2, data=None):  ##### TTF temperature dependence
    T_arr = np.split(T_arr2, 3)[0]
    # print(T_arr)
    f_start = params["f_start"]
    f_stop = params["f_stop"]
    w = 2 * np.pi * np.arange(f_start, f_stop, 1e5)
    C = params["C"]
    Rd = params["Rd"]
    # nu=params['nu']
    Tc = params["Tc"]
    L0 = params["L0"]
    Ls0 = params["Ls0"]
    RL0 = params["RL0"]

    RLs0 = Ls0 * RL0 / L0
    # D0=hbar/np.pi*nu*Tc
    w00 = 1 / np.sqrt(C * (L0 + Ls0))

    w0_list = []
    k0_list = []
    kex_list = []
    for j, T in enumerate(T_arr):
        RL = RL0 * fR1(T, Tc)
        RLs = RLs0 * fR1(T, Tc)
        # R=Rd*RL/(Rd+RL)
        L = L0 * fL1(T, Tc)
        Ls = Ls0 * fL1(T, Tc)
        # print(L)
        Rp = RL * Rd / (RL + Rd)
        Y = S11_series(w, C, L, Ls, Rp, RLs)
        p0 = [2 * np.pi * 4.6e9, 2 * np.pi * 1.5e6, 2 * np.pi * 1e6, 1, 0, 0, 0]
        popt, pcov = curve_fit(
            fit_S11,
            np.concatenate((w, w)),
            np.concatenate((np.real(Y), np.imag(Y))),
            p0=p0,
        )
        # fit Y to standard S11 ###
        ### get w0 and kint ###
        w0_list.append(popt[0])
        k0_list.append(popt[1] - popt[2])
        kex_list.append(popt[2])

    # w0=1/np.sqrt(C*(L*R**2/(w00**2*L**2+R**2)+Ls))
    w0 = np.array(w0_list)
    k0 = np.array(k0_list)
    kex = np.array(kex_list)

    if data is None:
        return np.concatenate((w0, k0, kex))
    else:
        return np.concatenate((w0, k0, kex)) - data


def MB_An_SeriesLC_direct(params, T_arr, data=None):  ##### MB theory, direct analytical
    # T_arr = np.split(T_arr2, 3)[0]
    # print(T_arr)
    try:
        Rex = params["Rex"]
    except:
        Rex = 50

    C = params["C"]
    Rd = params["Rd"]
    a = params["a"]
    Tc = params["Tc"]
    L0 = params["L0"]
    # RL0 = params["RL0"]

    # D0 = hbar / np.pi * nu * Tc
    D0 = a * KB * Tc
    # RL0= D0*L0/hbar/np.pi
    # RL0= D0*L0/hbar
    RL0 = D0 * L0 / hbar * np.pi / 2

    L = L0 * fL(T_arr, Tc, D0)
    RL = RL0 * fR(T_arr, Tc, D0)

    # w0 = 1 / np.sqrt(L * C)
    def Cav_imp(w):
        zl = 1 / (1 / (1j * w * L) + 1 / RL + 1 / Rd)
        zc = 1 / (1j * w * C) + zl
        return np.imag(zc)

    w0 = fsolve(Cav_imp, 1 / np.sqrt(L * C))

    ZL = 1 / (1 / (1j * w0 * L) + 1 / RL + 1 / Rd)

    k0 = np.real(ZL) / (np.imag(ZL) / w0)
    kex = Rex / (np.imag(ZL) / w0)

    if data is None:
        return np.concatenate((w0, k0, kex))
    else:
        return np.concatenate((w0, k0, kex)) - data


def MB_An_SeriesLC_direct_Cp(
    params, T_arr, data=None
):  ##### MB theory, direct analytical
    # T_arr = np.split(T_arr2, 3)[0]
    # print(T_arr)
    try:
        Rex = params["Rex"]
    except:
        Rex = 50

    # C = params["C"]
    Rd = params["Rd"]
    a = params["a"]
    Tc = params["Tc"]
    L0 = params["L0"]
    CLp = params["CLp"]
    w00 = params["w00"]
    # RL0 = params["RL0"]

    # D0 = hbar / np.pi * nu * Tc
    D0 = a * KB * Tc
    # RL0= D0*L0/hbar/np.pi
    # RL0= D0*L0/hbar
    RL0 = D0 * L0 / hbar * np.pi / 2

    L = L0 * fL(T_arr, Tc, D0)
    RL = RL0 * fR(T_arr, Tc, D0)
    w0 = w00 / np.sqrt(fL(T_arr, Tc, D0))

    # w0 = 1 / np.sqrt(L * C)
    # def Cav_imp(w):
    #     zl=1 / (1 / (1j * w * L) + 1 / RL + 1 / Rd+1j*w*CLp)
    #     zc=1/(1j*w*C)+zl
    #     return np.imag(zc)

    # w0=fsolve(Cav_imp, 1 / np.sqrt(L * C))

    ZL = 1 / (1 / (1j * w0 * L) + 1 / RL + 1 / Rd + 1j * w0 * CLp)

    k0 = np.real(ZL) / (np.imag(ZL) / w0)
    kex = Rex / (np.imag(ZL) / w0)

    if data is None:
        return np.concatenate((k0, kex))
    else:
        return np.concatenate((k0, kex)) - data


def TTF_An_SeriesLC_direct(params, T_arr, data=None):  ##### TTF, direct analytical
    # T_arr = np.split(T_arr2, 3)[0]
    # print(T_arr)
    try:
        Rex = params["Rex"]
    except:
        Rex = 50

    C = params["C"]
    Rd = params["Rd"]
    Tc = params["Tc"]
    L0 = params["L0"]
    RL0 = params["RL0"]

    L = L0 * fL1(T_arr, Tc)
    RL = RL0 * fR1(T_arr, Tc)

    def Cav_imp(w):
        zl = 1 / (1 / (1j * w * L) + 1 / RL + 1 / Rd)
        zc = 1 / (1j * w * C) + zl
        return np.imag(zc)

    w0 = fsolve(Cav_imp, 1 / np.sqrt(L * C))

    ZL = 1 / (1 / (1j * w0 * L) + 1 / RL + 1 / Rd)

    k0 = np.real(ZL) / (np.imag(ZL) / w0)
    kex = Rex / (np.imag(ZL) / w0)

    if data is None:
        return np.concatenate((w0, k0, kex))
    else:
        return np.concatenate((w0, k0, kex)) - data


def MB_An_Parallel_LC_Cc(params, T_arr, data=None):  ##### MB theory, direct analytical
    # T_arr = np.split(T_arr2, 3)[0]
    # print(T_arr)
    try:
        Rex = params["Rex"]
    except:
        Rex = 50

    Cc = params["Cc"]
    # RL0 = params["RL0"]
    C = params["C"]
    Rd = params["Rd"]
    a = params["a"]
    Tc = params["Tc"]
    L0 = params["L0"]

    # D0 = hbar / np.pi * nu * Tc
    D0 = a * KB * Tc
    RL0 = D0 * L0 / hbar / np.pi

    L = L0 * fL(T_arr, Tc, D0)
    RL = RL0 * fR(T_arr, Tc, D0)

    # w0 = 1 / np.sqrt(L * C)
    def Cav_imp(w):
        zp = 1 / (1j * w * C + 1 / RL + 1 / Rd + 1 / (1j * w * L))
        zc = 1 / (1j * w * Cc) + zp
        return np.imag(zc)

    # w0=fsolve(Cav_imp, 1 / np.sqrt(L * (C+Cc)))
    w0 = 1 / np.sqrt(L * (C + Cc))

    k0 = (1 / Rd + 1 / RL) * (1 / (C + Cc))
    kex = (w0**2 * Cc**2 * Rex) / (C + Cc)

    if data is None:
        return np.concatenate((w0, k0, kex))
    else:
        return np.concatenate((w0, k0, kex)) - data


def TTF_An_Parallel_LC_Cc(params, T_arr, data=None):  ##### MB theory, direct analytical
    # T_arr = np.split(T_arr2, 3)[0]
    # print(T_arr)
    try:
        Rex = params["Rex"]
    except:
        Rex = 50

    C = params["C"]
    Cc = params["Cc"]
    Rd = params["Rd"]

    Tc = params["Tc"]
    L0 = params["L0"]
    RL0 = params["RL0"]

    L = L0 * fL1(T_arr, Tc)
    RL = RL0 * fR1(T_arr, Tc)

    # w0 = 1 / np.sqrt(L * C)
    def Cav_imp(w):
        zp = 1 / (1j * w * C + 1 / RL + 1 / Rd + 1 / (1j * w * L))
        zc = 1 / (1j * w * Cc) + zp
        return np.imag(zc)

    # w0=fsolve(Cav_imp, 1 / np.sqrt(L * (C+Cc)))
    w0 = 1 / np.sqrt(L * (C + Cc))

    k0 = (1 / Rd + 1 / RL) * (1 / (C + Cc))
    kex = (w0**2 * Cc**2 * Rex) / (C + Cc)

    if data is None:
        return np.concatenate((w0, k0, kex))
    else:
        return np.concatenate((w0, k0, kex)) - data


def Zero_T(params, *args):
    w00, kex0, C = args
    L0, Ls0 = params
    F1 = 1 / np.sqrt(C * (L0 + Ls0)) - w00
    # F2=w00**2*L0**2/(Rd*L0)-k00
    F2 = (w00**2 * Ls0**2 * 50) / (L0 * (w00**2 * Ls0**2 + 50**2)) - kex0

    return [F1, F2]
